/*
 * Copyright (C) 2003 Trevor van Bremen
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1,
 * or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; see the file COPYING.LIB.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef	VBISAM_INCL	/* avoid multiple include problems */
#define	VBISAM_INCL

#include <sys/types.h>

#define	VBISAM_USE_LONG_LONG	/* Note hack for now */

#ifdef	VBISAM_USE_LONG_LONG
#define	vbisam_off_t	long long
#elif defined(VBISAM_USE_OFF_T)
#define	vbisam_off_t	off_t
#else
#define	vbisam_off_t	int
#endif

#ifdef	_MSC_VER
#ifdef VBISAM_LIB
#ifdef __cplusplus
#define DLL_EXPIMP extern "C" __declspec(dllexport)
#else
#define DLL_EXPIMP extern __declspec(dllexport)
#endif
#else
#ifdef __cplusplus
#define DLL_EXPIMP extern "C" __declspec(dllimport)
#else
#define DLL_EXPIMP extern __declspec(dllimport)
#endif
#endif
#else
#ifdef __cplusplus
#define DLL_EXPIMP extern "C"
#else
#define DLL_EXPIMP extern
#endif
#endif

#include	<string.h>
#include	<vbdecimal.h>

#define	CHARTYPE	0
#define	INTTYPE		1
#define	LONGTYPE	2
#define	DOUBLETYPE	3
#define	FLOATTYPE	4
#define	QUADTYPE	5

#define	CHARSIZE	1
#define	INTSIZE		2
#define	LONGSIZE	4
#define	FLOATSIZE	(sizeof(float))
#define	DOUBLESIZE	(sizeof(double))

#define	ISDESC		0x80

#define	BYTEMASK	0xFF	/* Mask for one byte */
#define	BYTESHFT	8		/* Shift for one byte */

#define	ISFIRST		0	/* Position to first row */
#define	ISLAST		1	/* Position to last row */
#define	ISNEXT		2	/* Position to next row */
#define	ISPREV		3	/* Position to previous row */
#define	ISCURR		4	/* Position to current row */
#define	ISEQUAL		5	/* Position to equal value */
#define	ISGREAT		6	/* Position to greater value */
#define	ISGTEQ		7	/* Position to >= value */

/* isread() locking modes */
#define	ISLOCK		0x100	/* Row lock */
#define	ISSKIPLOCK	0x200	/* Skip row even if locked */
#define	ISWAIT		0x400	/* Wait for row lock */
#define	ISLCKW		ISLOCK | ISWAIT

/* isstart() lock modes */
#define	ISKEEPLOCK	0x800	/* Keep rec lock in autolk mode */

/* isopen(), isbuild() lock modes */
#define	ISAUTOLOCK	0x200	/* Automatic row lock */
#define	ISMANULOCK	0x400	/* Manual row lock */
#define	ISEXCLLOCK	0x800	/* Exclusive isam file lock */

/* isopen(), isbuild() file types */
#define	ISINPUT		0	/* Open for input only */
#define	ISOUTPUT	1	/* Open for output only */
#define	ISINOUT		2	/* Open for input and output */
#define	ISTRANS		4	/* Open for transaction proc */
#define	ISNOLOG		8	/* No loggin for this file */
#define	ISVARLEN	0x10	/* Variable length rows */
#define	ISFIXLEN	0x0	/* (Non-flag) Fixed length rows only */

/* audit trail mode parameters */
#define	AUDSETNAME	0	/* Set new audit trail name */
#define	AUDGETNAME	1	/* Get audit trail name */
#define	AUDSTART	2	/* Start audit trail */
#define	AUDSTOP		3	/* Stop audit trail */
#define	AUDINFO		4	/* Audit trail running */

#define	VB_MAX_KEYLEN	511	/* BUG - FIXME! Maximum number of bytes in a key */
#define	NPARTS		8	/* Maximum number of key parts */

struct	keypart {
	short	kp_start;	/* Starting byte of key part */
	short	kp_leng;	/* Length in bytes */
	short	kp_type;	/* Type of key part (include ISDESC as needed) */
};

struct keydesc {
	short			k_flags;		/* Flags (Compression) */
	short			k_nparts;		/* Number of parts in this key */
	struct keypart	k_part[NPARTS];	/* Each key part */
	short			k_len;			/* Length of entire uncompressed key */
	vbisam_off_t	k_rootnode;		/* Pointer to rootnode */
#ifdef __cplusplus
	keydesc() { memset(this, 0, sizeof(keydesc)); }
#endif
};

#define	k_start	k_part[0].kp_start
#define	k_leng	k_part[0].kp_leng
#define	k_type	k_part[0].kp_type

/* Possible values for iFlags */
#define	ISNODUPS	0x00	/* No duplicates allowed */
#define	ISDUPS		0x01	/* Duplicates allowed */
#define	DCOMPRESS	0x02	/* Duplicate compression */
#define	LCOMPRESS	0x04	/* Leading compression */
#define	TCOMPRESS	0x08	/* Trailing compression */
#define	COMPRESS	0x0e	/* All compression */
#define	ISCLUSTER	0x00

struct	dictinfo
{
	short			di_nkeys;	/* Number of keys defined (msb set if VARLEN) */
	short			di_recsize;	/* Maximum data row length */
	short			di_idxsize;	/* Number of bytes in an index node */
	vbisam_off_t 	di_nrecords;/* Number of rows in data file */
};

/* Possible error return values */
#define	EDUPL		100	/* Duplicate row */
#ifndef ENOTOPEN
#define	ENOTOPEN	101	/* File not open */
#endif
#define	EBADARG		102	/* Illegal argument */
#define	EBADKEY		103	/* Illegal key desc */
#define	ETOOMANY	104	/* Too many files open */
#define	EBADFILE	105	/* Bad isam file format */
#define	ENOTEXCL	106	/* Non-exclusive access */
#ifndef	ELOCKED
#define	ELOCKED		107	/* Row locked */
#endif
#define	EKEXISTS	108	/* Key already exists */
#define	EPRIMKEY	109	/* Is primary key */
#define	EENDFILE	110	/* End/begin of file */
#ifndef	ENOREC
#define	ENOREC		111	/* No row found */
#endif
#define	ENOCURR		112	/* No current row */
#define	EFLOCKED	113	/* File locked */
#define	EFNAME		114	/* File name too long */
#define	EBADMEM		116	/* Can't alloc memory */
#define	ELOGREAD	118	/* Cannot read log rec */
#define	EBADLOG		119	/* Bad log row */
#define	ELOGOPEN	120	/* Cannot open log file */
#define	ELOGWRIT	121	/* Cannot write log rec */
#define	ENOTRANS	122	/* No transaction */
#define	ENOBEGIN	124	/* No begin work yet */
#define	ENOPRIM		127	/* No primary key */
#define	ENOLOG		128	/* No logging */
#define	ENOFREE		131	/* No free disk space */
#define	EROWSIZE	132	/* Row size too short / long */
#define	EAUDIT		133	/* Audit trail exists */
#define	ENOLOCKS	134	/* No more locks */
#define	EDEADLOK	143	/* Deadlock avoidance */
#define	ENOMANU		153	/* Must be in ISMANULOCK mode */
#define	EINTERUPT	157	/* Interrupted isam call */
#define	EBADFORMAT	171	/* Locking or NODESIZE change */

DLL_EXPIMP int	iserrno;	/* Isam error return code */
DLL_EXPIMP int	iserrio;	/* NOT used with VBISAM */
DLL_EXPIMP int	isreclen;	/* Used for varlen tables */
DLL_EXPIMP int	isrecnum;	/* Current row number */

struct	audhead
{
	char	au_type[2];		/* Audit row type aa,dd,rr,ww */
	char	au_time[4];		/* Audit date-time */
	char	au_procid[2];	/* Process id number */
	char	au_userid[2];	/* User id number */
	char	au_recnum[4];	/* Row number - RXW Set to 4 */
	char	au_reclen[2];	/* audit row length beyond header */
};

/* Number of bytes in audit header */
#define	AUDHEADSIZE	14
/* VARLEN num of bytes in audit header */
#define	VAUDHEADSIZE	16

/* Prototypes for file manipulation functions */
DLL_EXPIMP int	isaddindex(const int ihandle, struct keydesc * pskeydesc);
DLL_EXPIMP int	isaudit(const int ihandle, char * pcfilename, int imode);
DLL_EXPIMP int	isbegin(void);
DLL_EXPIMP int	isbuild(const char * pcfilename, const int imaxrowlength, struct keydesc * pskey, int imode);
DLL_EXPIMP int	ischeck(const char * pcfile);
DLL_EXPIMP int	iscleanup(void);
DLL_EXPIMP int	isclose(const int ihandle);
DLL_EXPIMP int	iscluster(const int ihandle, struct keydesc * pskeydesc);
DLL_EXPIMP int	iscommit(void);
DLL_EXPIMP int	isdelcurr(const int ihandle);
DLL_EXPIMP int	isdelete(const int ihandle, char * pcrow);
DLL_EXPIMP int	isdelindex(const int ihandle, struct keydesc * pskeydesc);
DLL_EXPIMP int	isdelrec(const int ihandle, vbisam_off_t trownumber);
DLL_EXPIMP int	iserase(char * pcfilename);
DLL_EXPIMP int	isflush(const int ihandle);
DLL_EXPIMP int	isfullclose(const int ihandle);
DLL_EXPIMP int	isindexinfo(const int ihandle, void * pskeydesc, const int ikeynumber);
DLL_EXPIMP int	islock(const int ihandle);
DLL_EXPIMP int	islogclose(void);
DLL_EXPIMP int	islogopen(const char * pcfilename);
DLL_EXPIMP int	isopen(const char * pcfilename, int imode);
DLL_EXPIMP int	isread(const int ihandle, char * pcrow, int imode);
DLL_EXPIMP int	isrecover(void);
DLL_EXPIMP int	isrelcurr(const int ihandle);
DLL_EXPIMP int	isrelease(const int ihandle);
DLL_EXPIMP int	isrelrec(const int ihandle, const vbisam_off_t trownumber);
DLL_EXPIMP int	isrename(char * pcoldname, char * pcnewname);
DLL_EXPIMP int	isrewcurr(const int ihandle, char * pcrow);
DLL_EXPIMP int	isrewrec(const int ihandle, const vbisam_off_t trownumber, char * pcrow);
DLL_EXPIMP int	isrewrite(const int ihandle, char * pcrow);
DLL_EXPIMP int	isrollback(void);
DLL_EXPIMP int	issetcollate(const int ihandle, const unsigned char * collating_sequence);
DLL_EXPIMP int	issetunique(const int ihandle, const vbisam_off_t tuniqueid);
DLL_EXPIMP int	isstart(const int ihandle, struct keydesc * pskeydesc, int ilength, char * pcrow, int imode);
DLL_EXPIMP int	isuniqueid(const int ihandle, vbisam_off_t * ptuniqueid);
DLL_EXPIMP int	isunlock(const int ihandle);
DLL_EXPIMP int	iswrcurr(const int ihandle, char * pcrow);
DLL_EXPIMP int	iswrite(const int ihandle, char * pcrow);
DLL_EXPIMP char	* isdi_name(const int ihandle);
DLL_EXPIMP int	isdi_datlen(const int ihandle);
DLL_EXPIMP int	isdi_curidx(const int ihandle);
DLL_EXPIMP int	isdi_idxfd(const int ihandle);
DLL_EXPIMP int	isdi_datfd(const int ihandle);
DLL_EXPIMP struct keydesc * isdi_kdsc(const int ihandle);

DLL_EXPIMP void	ldchar(char * pcsource, int ilength, char * pcdestination);
DLL_EXPIMP void	stchar(char * pcsource, char * pcdestination, int ilength);
DLL_EXPIMP int	ldint(void * pclocation);
DLL_EXPIMP void	stint(int ivalue, void * pclocation);
DLL_EXPIMP int	ldlong(void * pclocation);
DLL_EXPIMP void	stlong(int lvalue, void * pclocation);
/* RXW
DLL_EXPIMP vbisam_off_t	ldquad(void * pclocation);
DLL_EXPIMP void	stquad(vbisam_off_t tvalue, void * pclocation);
*/
DLL_EXPIMP double	ldfloat(void * pclocation);
DLL_EXPIMP void		stfloat(double dsource, void * pcdestination);
DLL_EXPIMP double	ldfltnull(void * pclocation, short * pinullflag);
DLL_EXPIMP void		stfltnull(double dsource, void * pcdestination, int inullflag);
DLL_EXPIMP double	lddbl(void * pclocation);
DLL_EXPIMP void		stdbl(double dsource, void * pcdestination);
DLL_EXPIMP double	lddblnull(void * pclocation, short * pinullflag);
DLL_EXPIMP void		stdblnull(double dsource, void * pcdestination, int inullflag);
DLL_EXPIMP int		lddecimal(unsigned char * cp, int len, dec_t * dp);
DLL_EXPIMP void		stdecimal(dec_t * dp, unsigned char * cp, int len);

#endif	/* VBISAM_INCL */
